! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                        Test for MPI_Barrier()
! 
! Thie code tests the functionality of the MPI_Barrier() function.  This
! MPI function is supposed to cause all processes in a given
! communicator to stop at a barrier until all others arrive at the same
! barrier. 
! 
! The idea of this test is for all nodes but one (which we will call the
! root) to immediately enter the barrier, and when they exit the
! barrier, they all send  a message to the root.  Before it enters the
! barrier, the root probes for messages from the other nodes.  If
! MPI_Barrier() is working properly, these probes (using the MPI
! function MPI_Iprobe()) should not find any messages.  If a message is
! found, then one of the non-root nodes must have passed the barrier.
! This is an error since the root has not yet entered the barrier.
! 
! If the root process does not find any messages after a fixed number of
! probes (the number of probes is fixed by the compiler constant
! MAX_PROBE), then it enters the barrier.  After exiting the barrier,
! the root process receives all the messages.
! 
! This test may be run in any type of communicator with any number of
! processes participating.
! 
! The default behaviour is to loop over the communicator types and sizes
! specified in the files defs.h and mpitest_cfgf.h .  See the
! MPITEST README file for details.
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    COMM_SIZE
!                               number of ranks in the current communicator
      LOGICAL    INTER_FLAG
!                               used to test if intercommunicator
      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    FLAG
!                               did the Iprobe find a message ?
      INTEGER    MAX_PROBE
      INTEGER    PROBE_COUNT
!                               the current number of completed probes
      INTEGER    ROOT
!                               the current checker for post-barrier messages
      INTEGER    COMM
!                               MPI communicator under test
      INTEGER    I, J

      INTEGER    BUFFER
      INTEGER    STATUS(MPI_STATUS_SIZE)
      INTEGER    SIZE
      INTEGER    POST_BARRIER_TAG

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 99   FORMAT(A,INT_FMT)
 179  FORMAT(A,I6,A,I6,A,I6,A,I6)

      MAX_PROBE=9000
      POST_BARRIER_TAG=0
!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Barrier()'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF


!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Get number of communicators to test
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
      IF (MPITEST_ME .EQ. 0) THEN
        WRITE (INFOBUF,99)'COMM_COUNT=',COMM_COUNT
        CALL MPITEST_MESSAGE (MPITEST_INFO2, INFOBUF)
      END IF

!
!     Loop through communicators
!
      DO 100 I=1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(I, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(I, COMM_TYPE)

        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)

!       Skip if intercommunicator

        IF (COMM .NE. MPI_COMM_NULL) THEN
          CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
          IF (INTER_FLAG .EQV. .TRUE.) GO TO 201
        END IF

!
!       Loop over different processes being the root
!
        DO 200 ROOT=0, COMM_SIZE-1
!         Print an informational message
          IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
            WRITE (INFOBUF,179) 'commindex', COMM_INDEX, ', commsize', &
                COMM_SIZE, ', commtype', COMM_TYPE, ', root', ROOT
            CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
          END IF

!
!         Only execute the test if in the current communicator
          IF (COMM .NE. MPI_COMM_NULL) THEN
            LOOP_CNT = LOOP_CNT + 1
!
!           The root node probes, then enters the barrier, then 
!           receives messages
!
            IF (MPITEST_CURRENT_RANK .EQ. ROOT) THEN
              FLAG = 0
              PROBE_COUNT = 0
              DO 600 J=0, MAX_PROBE
!
!               Look for post-barrier messages from any other process
!
                CALL MPI_IPROBE(MPI_ANY_SOURCE, POST_BARRIER_TAG,  &
                                  COMM, FLAG, STATUS, IERR)
                IF (IERR .NE. MPI_SUCCESS) THEN
                  WRITE(INFOBUF,99) 'MPI_IPROBE() returned', IERR
                  CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                  INFOBUF=' '
                  CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
                  CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
                END IF
                IF (FLAG .NE. 0) THEN
                  PROBE_COUNT = MAX_PROBE
                END IF
 600          CONTINUE
!
!             If flag is true, then a post-barrier message
!             was received pre-barrier, so the test fails 
!
              IF (FLAG .NE. 0) THEN
                FAIL = FAIL + 1
                INFOBUF = 'Received a post-barrier message '// &
                          'before entering barrier'
                CALL MPITEST_MESSAGE (MPITEST_NONFATAL, INFOBUF)
              END IF
                  
              CALL MPI_BARRIER(COMM, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                WRITE(INFOBUF,99) 'MPI_BARRIER() returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
              END IF
                  
!             Root process receives post-barrier messages
              IF (COMM_SIZE .GT. 1) THEN
                DO 400 J=0, COMM_SIZE-2
                  CALL MPI_RECV(BUFFER, 0, MPI_INTEGER,  &
                         MPI_ANY_SOURCE, POST_BARRIER_TAG, COMM, &
                         STATUS, IERR)
                  IF (IERR .NE. MPI_SUCCESS) THEN
                    WRITE(INFOBUF,99) 'MPI_RECV() returned', IERR
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
                    CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
                  END IF
 400            CONTINUE
              END IF
!
!             Non-root processes enter barrier then send message to root
!
            ELSE
              CALL MPI_BARRIER(COMM, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                WRITE(INFOBUF,99) 'MPI_BARRIER() returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
              END IF
              CALL MPI_SEND( BUFFER, 0, MPI_INTEGER, ROOT, &
                    POST_BARRIER_TAG, COMM, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                WRITE(INFOBUF,99) 'MPI_SEND() returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
              END IF
            END IF
          END IF
      
 200    CONTINUE
 201    CONTINUE
      
      
        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)
 100  CONTINUE

!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END  
  
  
