! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!      Test for MPI_Isend():  ROOT send to ALL version with BYTE offsets
!           and MPI_Irecv()
!
! This is a modified version of the  MPI_Isend  INTER-COMMUNICATOR version
! of the SEND FROM ROOT TO ALL. The program uses only the CHARACTER data type.
! The charater records are then offset by one to three bytes for both send and
! receive as a test for non-word address boundary conditions.
! 
! The program selects each node, in turn, to be the root.  The ROOT SENDS to
! each of the other nodes.  In this test ROOT does NOT send to itself.
! 
! This test initializes the send buffer with the root's rank in the
! communicator (or an appropriate value for the non-integer types.)
! Once the receiving nodes have completed their message receive, they check to
! make sure the current root's rank is in the received buffer.
! 
! This test may be run in any communicator, with any data type, and with
! any non-negative message length.
! 
! The MPITEST environment provides looping over communicator size and
! type, message length, and data type.  The properties of the loops are
! encoded in configuration arrays in the file config.h .  See the
! MPITEST README for further details.
! 
! **********************************************************************
#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM TEST
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    COMM_SIZE
!                               number of ranks in the current communicator
      INTEGER    TYPE_COUNT
!                               number of data types to test
      INTEGER    TEST_TYPE
!                               index of current data type
      INTEGER    LENGTH_COUNT
!                               number of data lengths to test
      INTEGER    MAX_LENGTH
!                               max message length to test
      INTEGER    MAX_BYTE_LENGTH
!                               max byte length to test
      INTEGER    MAX_BUFF_LENGTH
!                               max length that will fit in buffers
      INTEGER    BYTE_LENGTH
!                               current byte length to test
      INTEGER    LENGTH
!                               current length to test 

      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    LOOP_FAIL
!                               counts number of failures in 1 loop
      INTEGER    ERROR, ERROR2
!                               number of data errors found in buffer
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ROOT
!                               the current root
      INTEGER    COMM
!                               MPI communicator under test
      INTEGER    COUNT_LEN
!                               Length of received message from MPI_GET_COUNT
      LOGICAL    PRINT_NODE
!                              Is node to print informational messages

      INTEGER    I, J, K, L, M

      MPITEST_BUF_TYPE SEND_BUFFER(MAX_BUFF_SIZE+5)
      MPITEST_BUF_TYPE RECV_BUFFER(MAX_BUFF_SIZE+5)
      INTEGER    ERRSIZE

      INTEGER    SEND_STAT(MPI_STATUS_SIZE)
      INTEGER    RECV_STAT(MPI_STATUS_SIZE)
      INTEGER    SEND_REQ
      INTEGER    RECV_REQ

      INTEGER    OFFSET
      INTEGER    SEND_OFFSET
      INTEGER    RECV_OFFSET

      MPITEST_BUF_TYPE VALUE

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 98   FORMAT(A,INT_FMT,A,INT_FMT)
 99   FORMAT(A,INT_FMT)
 179  FORMAT(A,I6,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)
 189  FORMAT(INT_FMT,A,A,I6,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)

!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Isend_off'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Get number of communicators, data types and message lengths to test
!
      CALL MPITEST_NUM_MESSAGE_LENGTHS(LENGTH_COUNT)

!
!     Get max message length for test
!
      CALL MPITEST_GET_MAX_MESSAGE_LENGTH(MAX_BYTE_LENGTH)

      COMM = MPI_COMM_WORLD
      IF (MPITEST_CURRENT_RANK .EQ. 0) PRINT_NODE = .TRUE.

!
!     We will just test characters in this test, so force it here
!
      MPITEST_TYPES(1) = MPITEST_CHARACTER
      J = 1
      CALL MPITEST_GET_DATATYPE(J, TEST_TYPE)
!
!     Convert the number of bytes in the maximum length message
!     into the number of elements of the current type 
      CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, MAX_BYTE_LENGTH, &
              MAX_LENGTH)
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                 MAX_BUFF_SIZE*MPITEST_BUF_EXTENT, MAX_BUFF_LENGTH)

      IF (MAX_LENGTH .GE. (MAX_BUFF_LENGTH - 5)) &
           MAX_LENGTH = MAX_BUFF_LENGTH - 5


!
!     Loop through data lengths
!
      DO 200 K=1,LENGTH_COUNT
        CALL MPITEST_GET_MESSAGE_LENGTH(K, BYTE_LENGTH)
        CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, BYTE_LENGTH, &
            LENGTH)
        IF (LENGTH .GT. MAX_LENGTH) LENGTH = MAX_LENGTH

!
!       Loop through offsets
!
        DO 300 OFFSET=1,4
          SEND_OFFSET = OFFSET
          RECV_OFFSET = 5 - OFFSET

!
!       Loop over different processes being the root
!

          DO 400 ROOT=0,MPITEST_NUMP-1
!           Print an informational message
            IF (PRINT_NODE) THEN
              INFOBUF=' '
              WRITE (INFOBUF,179)  &
                'commindex', COMM_INDEX, ', commsize', &
                COMM_SIZE, ', commtype', COMM_TYPE, ', root', ROOT, &
                ', datatype', TEST_TYPE, ', length',  LENGTH
              CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
            END IF

!
!           All but root on receive side post receives
!
            IF (MPITEST_CURRENT_RANK .NE. ROOT) THEN
!
!             Initialize expected receive data
!
              CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1, &
                TEST_TYPE)
              CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH+1, VALUE, &
                RECV_BUFFER(RECV_OFFSET))
              LOOP_CNT = LOOP_CNT + 1
              LOOP_FAIL = 0


              CALL MPI_IRECV(RECV_BUFFER(RECV_OFFSET), LENGTH+1, &
                MPITEST_MPI_DATATYPES(TEST_TYPE), ROOT, ROOT, &
                COMM, RECV_REQ, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                INFOBUF=' '
                WRITE(INFOBUF,99) 'MPI_RECV() returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                LOOP_FAIL = LOOP_FAIL + 1
              END IF

              CALL MPI_WAIT(RECV_REQ, RECV_STAT, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                INFOBUF=' '
                WRITE(INFOBUF,99) 'MPI_WAIT(IRECV) returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                IF (RECV_STAT(MPI_ERROR) .NE. MPI_SUCCESS) THEN
                  INFOBUF=' '
                  CALL MPI_ERROR_STRING(RECV_STAT(MPI_ERROR), &
                        INFOBUF, ERRSIZE, ERR)
                  CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                END IF
                FAIL = FAIL + 1
              END IF

!
!             Check data for errors and overflow
!
              CALL MPITEST_DATATEMPLATE_INIT(VALUE, ROOT, &
                TEST_TYPE)
              CALL MPITEST_BUFFER_ERRORS(TEST_TYPE, LENGTH, VALUE, &
                RECV_BUFFER(RECV_OFFSET), ERROR)
              CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1, &
                TEST_TYPE)
              CALL MPITEST_BUFFER_ERRORS_OV(TEST_TYPE, LENGTH, &
                VALUE, RECV_BUFFER(RECV_OFFSET), ERROR2)
              IF ((ERROR + ERROR2) .NE. 0) THEN
                INFOBUF=' '
                WRITE (INFOBUF,189) ERROR+ERROR2, &
                  ' errors in buffer, ', &
                  'commindex',COMM_INDEX,', commsize', COMM_SIZE, &
                  ', commtype', COMM_TYPE, ', root', ROOT, &
                  ', datatype', TEST_TYPE, ', length',  LENGTH
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                LOOP_FAIL = LOOP_FAIL + 1
              END IF
                
!
!             Call MPI_GET_COUNT and compare value with length received
!
              CALL MPI_GET_COUNT(RECV_STAT,  &
                  MPITEST_MPI_DATATYPES(TEST_TYPE), COUNT_LEN, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                INFOBUF=' '
                WRITE(INFOBUF,99) 'MPI_GET_COUNT() returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                LOOP_FAIL = LOOP_FAIL + 1
              END IF

              IF (LENGTH - COUNT_LEN .NE. 0) THEN
                INFOBUF=' '
                WRITE(INFOBUF,98) 'Unexpected receive length', &
                    COUNT_LEN, ', expected', LENGTH
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                LOOP_FAIL = LOOP_FAIL + 1
              END IF

!
!             Check source and tag values.
!
              IF (RECV_STAT(MPI_TAG) .NE. ROOT) THEN
                INFOBUF=' '
                WRITE(INFOBUF,98) 'Unexpected tag value', &
                  RECV_STAT(MPI_TAG), ', expected', ROOT
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                LOOP_FAIL = LOOP_FAIL + 1
              END IF

              IF (RECV_STAT(MPI_SOURCE) .NE. ROOT) THEN
                INFOBUF=' '
                WRITE(INFOBUF,98) 'Unexpected source value', &
                  RECV_STAT(MPI_SOURCE), ', expected', ROOT
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                LOOP_FAIL = LOOP_FAIL + 1
              END IF

              IF (LOOP_FAIL .NE. 0) FAIL = FAIL + 1
            END IF
!                Receiver

!
!           Root sends the message
!
            IF (MPITEST_CURRENT_RANK .EQ. ROOT) THEN

              CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1,  &
                TEST_TYPE)
              CALL MPITEST_INIT_BUFFER(TEST_TYPE, 4, &
                VALUE, SEND_BUFFER(1))

              CALL MPITEST_DATATEMPLATE_INIT(VALUE,  &
                MPITEST_CURRENT_RANK, TEST_TYPE)
              CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH+1, &
                VALUE, SEND_BUFFER(SEND_OFFSET))

              DO 500 M=0, MPITEST_NUMP-1
                IF (M .EQ. ROOT) GO TO 500
!                       Don't send to self
                LOOP_CNT = LOOP_CNT + 1

                CALL MPI_ISEND(SEND_BUFFER(SEND_OFFSET), LENGTH, &
                  MPITEST_MPI_DATATYPES(TEST_TYPE), M, ROOT, &
                  COMM, SEND_REQ, IERR)
                IF (IERR .NE. MPI_SUCCESS) THEN
                  INFOBUF=' '
                  WRITE(INFOBUF,99) 'MPI_ISEND() returned', IERR
                  CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                  INFOBUF=' '
                  CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
                  CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                  FAIL = FAIL + 1
                END IF

                CALL MPI_WAIT(SEND_REQ, SEND_STAT, IERR)
                IF (IERR .NE. MPI_SUCCESS) THEN
                  INFOBUF=' '
                  WRITE(INFOBUF,99) 'MPI_WAIT(ISEND) returned', IERR
                  CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                  INFOBUF=' '
                  CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
                  CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                  IF (SEND_STAT(MPI_ERROR) .NE. MPI_SUCCESS) THEN
                    INFOBUF=' '
                    CALL MPI_ERROR_STRING(SEND_STAT(MPI_ERROR), &
                          INFOBUF, ERRSIZE, ERR)
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                  END IF
                  FAIL = FAIL + 1
                END IF

 500          CONTINUE
!                         Isend to all receivers
            END IF
!                    Sender

            CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              INFOBUF=' '
              WRITE(INFOBUF,99) 'MPI_BARRIER() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF

 400      CONTINUE
!                     End root loop


 300    CONTINUE
!                   End data offset loop

 200  CONTINUE
!                 End data length loop
      

!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END
